"""Index of archived XLSForm file paths."""

import os

xlsforms_path = os.path.dirname(os.path.abspath(__file__))

amenities = f"{xlsforms_path}/amenities.xls"
buildings = f"{xlsforms_path}/buildings.xls"
camping = f"{xlsforms_path}/camping.xls"
cemeteries = f"{xlsforms_path}/cemeteries.xls"
education = f"{xlsforms_path}/education.xls"
health = f"{xlsforms_path}/health.xls"
highways = f"{xlsforms_path}/highways.xls"
historical = f"{xlsforms_path}/historical.xls"
hotspring = f"{xlsforms_path}/hotspring.xls"
landusage = f"{xlsforms_path}/landusage.xls"
landuse = f"{xlsforms_path}/landuse.xls"
municipality = f"{xlsforms_path}/municipality.xls"
nature = f"{xlsforms_path}/nature.xls"
places = f"{xlsforms_path}/places.xls"
religious = f"{xlsforms_path}/religious.xls"
solidwaste = f"{xlsforms_path}/solidwaste.xls"
toilets = f"{xlsforms_path}/toilets.xls"
towns = f"{xlsforms_path}/towns.xls"
transportation = f"{xlsforms_path}/transportation.xls"
waste_collection = f"{xlsforms_path}/waste_collection.xlsx"
wastedisposal = f"{xlsforms_path}/wastedisposal.xls"
waterpoints = f"{xlsforms_path}/waterpoints.xls"
waterways = f"{xlsforms_path}/waterways.xls"
