"""Index of XLSForm file paths."""

import os

xlsforms_path = os.path.dirname(os.path.abspath(__file__))

entities_registration = f"{xlsforms_path}/entities_registration.xls"
entities_registration_xml = f"{xlsforms_path}/entities_registration.xml"

buildings = f"{xlsforms_path}/buildings.xls"
